{# TEMPLATE VAR SETTINGS #}
{%- set url_root = pathto('', 1) %}
{%- if url_root == '#' %}{% set url_root = '' %}{% endif %}
{%- if not embedded and docstitle %}
  {%- set titlesuffix = " &mdash; "|safe + docstitle|e %}
{%- else %}
  {%- set titlesuffix = "" %}
{%- endif %}
{%- set lang_attr = 'en' if language == None else (language | replace('_', '-')) %}

<!DOCTYPE html>
<!--[if IE 8]><html class="no-js lt-ie9" lang="{{ lang_attr }}" > <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js" lang="{{ lang_attr }}" > <!--<![endif]-->
<head>
  <meta charset="utf-8">
  {{ metatags }}
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  {% block htmltitle %}
  <title>{{ title|striptags|e }}{{ titlesuffix }}</title>
  {% endblock %}

  {# FAVICON #}
  {% if favicon %}
    <link rel="shortcut icon" href="{{ pathto('_static/' + favicon, 1) }}"/>
  {% endif %}
  {# CANONICAL URL #}
  {% if theme_canonical_url %}
    <link rel="canonical" href="{{ theme_canonical_url }}{{ pagename }}.html"/>
  {% endif %}

  {# JAVASCRIPTS #}
  {%- block scripts %}
  <script type="text/javascript" src="{{ pathto('_static/js/modernizr.min.js', 1) }}"></script>
  {%- if not embedded %}
  {# XXX Sphinx 1.8.0 made this an external js-file, quick fix until we refactor the template to inherert more blocks directly from sphinx #}
    {% if sphinx_version >= "1.8.0" %}
      <script type="text/javascript" id="documentation_options" data-url_root="{{ pathto('', 1) }}" src="{{ pathto('_static/documentation_options.js', 1) }}"></script>
      {%- for scriptfile in script_files %}
        {{ js_tag(scriptfile) }}
      {%- endfor %}
    {% else %}
      <script type="text/javascript">
          var DOCUMENTATION_OPTIONS = {
              URL_ROOT:'{{ url_root }}',
              VERSION:'{{ release|e }}',
              LANGUAGE:'{{ language }}',
              COLLAPSE_INDEX:false,
              FILE_SUFFIX:'{{ '' if no_search_suffix else file_suffix }}',
              HAS_SOURCE:  {{ has_source|lower }},
              SOURCELINK_SUFFIX: '{{ sourcelink_suffix }}'
          };
      </script>
      {%- for scriptfile in script_files %}
        <script type="text/javascript" src="{{ pathto(scriptfile, 1) }}"></script>
      {%- endfor %}
    {% endif %}
    <script type="text/javascript" src="{{ pathto('_static/js/theme.js', 1) }}"></script>

    {# OPENSEARCH #}
    {%- if use_opensearch %}
    <link rel="search" type="application/opensearchdescription+xml"
          title="{% trans docstitle=docstitle|e %}Search within {{ docstitle }}{% endtrans %}"
          href="{{ pathto('_static/opensearch.xml', 1) }}"/>
    {%- endif %}
  {%- endif %}
  {%- endblock %}

  {# Ansible CCS additions #}
  <link href='https://fonts.googleapis.com/css?family=Lato:400,700|Roboto+Slab:400,700|Inconsolata:400,700' rel='stylesheet' type='text/css'>
  <link href='https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.0.3/css/font-awesome.min.css' rel='stylesheet' type='text/css'>

  {# CSS #}
  <link rel="stylesheet" href="{{ pathto('_static/' + style, 1) }}" type="text/css" />
  <link rel="stylesheet" href="{{ pathto('_static/ansible.css', 1) }}" type="text/css" />
  <link rel="stylesheet" href="{{ pathto('_static/pygments.css', 1) }}" type="text/css" />
  {%- for css in css_files %}
    {%- if css|attr("rel") %}
  <link rel="{{ css.rel }}" href="{{ pathto(css.filename, 1) }}" type="text/css"{% if css.title is not none %} title="{{ css.title }}"{% endif %} />
    {%- else %}
  <link rel="stylesheet" href="{{ pathto(css, 1) }}" type="text/css" />
    {%- endif %}
  {%- endfor %}

  {%- for cssfile in extra_css_files %}
    <link rel="stylesheet" href="{{ pathto(cssfile, 1) }}" type="text/css" />
  {%- endfor %}

  {%- block linktags %}
    {%- if hasdoc('about') %}
    <link rel="author" title="{{ _('About these documents') }}" href="{{ pathto('about') }}" />
    {%- endif %}
    {%- if hasdoc('genindex') %}
    <link rel="index" title="{{ _('Index') }}" href="{{ pathto('genindex') }}" />
    {%- endif %}
    {%- if hasdoc('search') %}
    <link rel="search" title="{{ _('Search') }}" href="{{ pathto('search') }}" />
    {%- endif %}
    {%- if hasdoc('copyright') %}
    <link rel="copyright" title="{{ _('Copyright') }}" href="{{ pathto('copyright') }}" />
    {%- endif %}
    {%- if next %}
    <link rel="next" title="{{ next.title|striptags|e }}" href="{{ next.link|e }}" />
    {%- endif %}
    {%- if prev %}
    <link rel="prev" title="{{ prev.title|striptags|e }}" href="{{ prev.link|e }}" />
    {%- endif %}
  {%- endblock %}
  {%- block extrahead %} {% include "ansible_extrahead.html" %} {% endblock %}
</head>

<body class="wy-body-for-nav">
  {% block extrabody %} {% include "ansible_extrabody.html" %} {% endblock %}
  <div class="wy-grid-for-nav">
    {# SIDE NAV, TOGGLES ON MOBILE #}
    <nav data-toggle="wy-nav-shift" class="wy-nav-side">
      <div class="wy-side-scroll">
        <div class="wy-side-nav-search" {% if theme_style_nav_header_background %} style="background: {{theme_style_nav_header_background}}" {% endif %}>
          {% block sidebartitle %}

          {% if logo and theme_logo_only %}
            <a href="{{ pathto(master_doc) }}">
          {% else %}
            <a href="{{ pathto(master_doc) }}" class="icon icon-home"> {{ project }}
          {% endif %}

          {% if logo %}
            {# Not strictly valid HTML, but it's the only way to display/scale
               it properly, without weird scripting or heaps of work
            #}
            <img src="{{ pathto('_static/' + logo, 1) }}" class="logo" alt="Logo"/>
          {% endif %}
          </a>

          {% if theme_display_version %}
            {%- set nav_version = version %}
            {% if READTHEDOCS and current_version %}
              {%- set nav_version = current_version %}
            {% endif %}
            {% if nav_version %}
              <div class="version">
                {{ nav_version }}
              </div>
            {% endif %}
          {% endif %}

          {% include "ansible_versions.html" %}
          {% include "ansible_searchbox.html" %}

          {% endblock %}
        </div>

        {% block navigation %}
        <div class="wy-menu wy-menu-vertical" data-spy="affix" role="navigation" aria-label="main navigation">
          {% block menu %}
            {#
              The singlehtml builder doesn't handle this toctree call when the
              toctree is empty. Skip building this for now.
            #}
            {% if 'singlehtml' not in builder %}
              {% set global_toc = toctree(maxdepth=theme_navigation_depth|int,
                                          collapse=theme_collapse_navigation|tobool,
                                          includehidden=theme_includehidden|tobool,
                                          titles_only=theme_titles_only|tobool) %}
            {% endif %}
            {% if global_toc %}
              {{ global_toc }}
            {% else %}
              <!-- Local TOC -->
              <div class="local-toc">{{ toc }}</div>
            {% endif %}
          {% endblock %}
        </div>
        {% endblock %}
         {% block extranav %}{% include "ansible_extranav.html" %}{% endblock %}
      </div>
    </nav>
  </div>

    <section data-toggle="wy-nav-shift" class="wy-nav-content-wrap">

      {# MOBILE NAV, TRIGGLES SIDE NAV ON TOGGLE #}
      <nav class="wy-nav-top" aria-label="top navigation">
        {% block mobile_nav %}
          <i data-toggle="wy-nav-top" class="fa fa-bars"></i>
          <a href="{{ pathto(master_doc) }}">{{ project }}</a>
        {% endblock %}
      </nav>


      <div class="wy-nav-content">
      {%- block content %}
        {% if theme_style_external_links|tobool %}
        <div class="rst-content style-external-links">
        {% else %}
        <div class="rst-content">
        {% endif %}
          {% include "breadcrumbs.html" %}
          {% include "ansible_banner.html" %}
          <div role="main" class="document" itemscope="itemscope" itemtype="http://schema.org/Article">
          {%- block document %}
           <div itemprop="articleBody">
            {% block body %}{% endblock %}
           </div>
           {% if self.comments()|trim %}
           <div class="articleComments">
            {% block comments %}{% endblock %}
           </div>
           {% endif%}
          </div>
          {%- endblock %}
          {% include "footer.html" %}
        </div>
      {%- endblock %}
      </div>

    </section>

  </div>

  <script type="text/javascript">
      jQuery(function () {
          {% if theme_sticky_navigation|tobool %}
          SphinxRtdTheme.Navigation.enableSticky();
          {% else %}
          SphinxRtdTheme.Navigation.enable();
          {% endif %}
      });
  </script>

  {# Do not conflict with RTD insertion of analytics script #}
  {% if not READTHEDOCS %}
    {% if theme_analytics_id %}
    <!-- Theme Analytics -->
    <script>
    (function(i,s,o,g,r,a,m){i['GoogleAnalyticsObject']=r;i[r]=i[r]||function(){
      (i[r].q=i[r].q||[]).push(arguments)},i[r].l=1*new Date();a=s.createElement(o),
      m=s.getElementsByTagName(o)[0];a.async=1;a.src=g;m.parentNode.insertBefore(a,m)
    })(window,document,'script','https://www.google-analytics.com/analytics.js','ga');

    ga('create', '{{ theme_analytics_id }}', 'auto');
    ga('send', 'pageview');
    </script>

    {% endif %}
  {% endif %}

  {%- block footer %} {% include "ansible_extrafooter.html" %} {% endblock %}

</body>
</html>
