{%- if builder != 'singlehtml' %}
<div role="search">
  <form id="rtd-search-form" class="wy-form" action="{{ pathto('search') }}"
    <input type="text" name="q" placeholder="{{ _('   Search') }}" />
    <input type="hidden" name="check_keywords" value="yes" />
    <input type="hidden" name="area" value="default" />
  </form>
</div>
{%- endif %}
