.. _@{ slug }@:

Plugins Maintained by the @{ maintainers }@
``````````````````````````@{ '`' * maintainers | length }@

.. toctree:: :maxdepth: 1

{% for module in modules | sort %}
  @{ module }@{% if module_info[module]['deprecated'] %} **(D)**{% endif %} - @{ module_info[module]['doc']['short_description'] }@ <plugins/@{ module_info[module]['primary_category'] }@/@{ module }@>
{% endfor %}

.. note::
    - **(D)**: This marks a plugin as deprecated, which means a plugin is kept for backwards compatibility but usage is discouraged.
      The plugin documentation details page may explain more about this rationale.

